#ifndef __FILE_OBJECT_H_INCLUDED__
#define __FILE_OBJECT_H_INCLUDED__

#include "block.h"

typedef struct _FileObject *FileObject;

typedef enum {
  FILE_OBJECT_TYPE_NONE,
  FILE_OBJECT_TYPE_DIR,
  FILE_OBJECT_TYPE_FILE,
} file_object_type_t;

/******** for FileObject operation ********/

BlockPool FileObject_GetPool(FileObject object);
file_object_type_t FileObject_GetType(FileObject object);
file_object_type_t FileObject_SetType(FileObject object, file_object_type_t type);
unsigned int FileObject_GetFlags(FileObject object);
unsigned int FileObject_SetFlags(FileObject object, unsigned int flags);
int FileObject_GetReference(FileObject object);
int FileObject_SetReference(FileObject object, int reference);
int FileObject_ClearReference(FileObject object);
int FileObject_AddReference(FileObject object);
int FileObject_DelReference(FileObject object);
FileObject FileObject_Create(BlockPool pool, file_object_type_t type);
FileObject FileObject_Destroy(FileObject object);

/******** for FileData operation ********/

int FileObject_DataGetSize(FileObject object);
int FileObject_DataRead(FileObject object, int offset, int size, void *buffer);
int FileObject_DataWrite(FileObject object, int offset, int size, const void *buffer);
int FileObject_DataCompare(FileObject object, int offset, int size, const void *buffer);
int FileObject_DataClear(FileObject object);
int FileObject_DataExtend(FileObject object, int size);
int FileObject_DataReduce(FileObject object, int size);
int FileObject_DataResize(FileObject object, int size);

#endif
