#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fuse/fuse.h>

#include "file_node.h"
#include "file_object.h"
#include "file_descriptor.h"
#include "file.h"
#include "fs.h"

static struct fs_info *get_fs_info()
{
  struct fuse_context *context;
  struct fs_info *fs;

  context = fuse_get_context();
  fs = context->private_data;
  return fs;
}

static uid_t get_user_id()
{
  return fuse_get_context()->uid;
}

static gid_t get_group_id()
{
  return fuse_get_context()->gid;
}

#if 0
static BlockPool get_pool()
{
  return get_fs_info()->pool;
}
#endif

static FileNode get_mountpoint_node()
{
  return get_fs_info()->mountpoint;
}

#if 0
static FileNode get_root_node()
{
  return get_fs_info()->root;
}
#endif

/*
 * ѥ̾եΥΡɤƵŪ˸롥
 * ĤäեͤǡΥե뤬¸ߤǥ쥯ȥ
 * dir ֤ե̾ˤϥǥ쥯ȥ̾Ǥ롥
 * Ĥʤä NULL ֤Υե(orǥ쥯ȥ)
 * ¸ߤ٤ǥ쥯ȥ dir ֤
 * nodenameˤϡѥ̾Υǥ쥯ȥ̾ñΥե֤̾
 * .פ..פˤбʤ(פʤΤ)
 * root ˥ǥ쥯ȥϤȡΥǥ쥯ȥˤƸ롥
 * root  NULL Ϥȡ롼ȥǥ쥯ȥˤƸ롥
 */
static FileNode find_node_proc(char *path, FileNode root, FileNode *dir, char **nodename)
{
  FileNode node;
  char *p, *next;

  p = strchr(path, '/');
  if (p) {
    *p = '\0';
    next = p + 1;
    if (next[0] == '\0')
      next = NULL;
  } else {
    next = NULL;
  }

  if (!root) root = get_mountpoint_node();
  node = Dir_SearchNode(root, path);

  if (node && next)
    return find_node_proc(next, node, dir, nodename);

  /*
   * ѥ̾Υǥ쥯ȥ꤬¸ߤʤ硤ޤǤΥǥ쥯ȥ꤬
   * dirä֤äƤޤХ
   */

  if (dir) *dir = root;
  if (nodename) *nodename = path;

  return node;
}

static FileNode find_node(const char *path, FileNode root, FileNode *dir, char **nodename)
{
  FileNode node;
  static char *pathbuf = NULL;
  static int size = 0;
  int len;

  len = strlen(path);

  if (!size || (size < len + 1)) {
    size = len + 1;
    if (pathbuf) free(pathbuf);
    pathbuf = malloc(size);
  }

  strcpy(pathbuf, path);
  node = find_node_proc(pathbuf, root, dir, nodename);

  return node;
}

static int get_stat(FileNode node, struct stat *st)
{
  FileObject object = FileNode_GetObject(node);

  memset(st, 0, sizeof(*st));

  st->st_dev = 0; /* ignored */
  st->st_ino = 0; /* ignored */

  st->st_mode = FileNode_GetFlags(node) & ACCESSPERMS;

  switch (FileObject_GetType(object)) {
  case FILE_OBJECT_TYPE_DIR:
    st->st_mode |= S_IFDIR;
    break;
  case FILE_OBJECT_TYPE_FILE:
  case FILE_OBJECT_TYPE_NONE:
  default:
    st->st_mode |= S_IFREG;
    break;
  };

  /*
   * եξϣǥ쥯ȥξݻƤ
   * ֥ǥ쥯ȥοܣ (ܣϤ餯 .  .. Τ֤)֤褦
   */
  st->st_nlink = 1;
  if (FileObject_GetType(object) == FILE_OBJECT_TYPE_DIR)
    st->st_nlink = Dir_GetEntryNumberByType(node, FILE_OBJECT_TYPE_DIR) + 2;

  st->st_uid = FileNode_GetUserID(node);
  st->st_gid = FileNode_GetGroupID(node);
  st->st_rdev = -1;

  st->st_atime = FileNode_GetAccessTime(node);
  st->st_mtime = FileNode_GetModifiedTime(node);
  st->st_ctime = FileNode_GetChangedTime(node);

  st->st_size = FileObject_DataGetSize(object);

  st->st_blocks  = 0;
  st->st_blksize = 0; /* ignored */

#if 0 /* FreeBSDǤ¸ߤ뤬LinuxǤ¸ߤʤߤ */
  st->st_flags   = 0;
  st->st_gen     = 0;
  st->st_lspare  = 0;

  st->st_birthtime = 0;
#endif

  return 0;
}

static int check_permit(FileNode node, uid_t uid, gid_t gid, int access)
{
  mode_t mode, check = 0;

  if (uid == 0) /* super user */
    return 1;

  if ((access & R_OK)) check |= S_IRUSR | S_IRGRP | S_IROTH;
  if ((access & W_OK)) check |= S_IWUSR | S_IWGRP | S_IWOTH;
  if ((access & X_OK)) check |= S_IXUSR | S_IXGRP | S_IXOTH;

  if (FileNode_GetUserID(node) == uid) { /* owner */
    check &= S_IRWXU;
  } else if (FileNode_GetGroupID(node) == gid) { /* same group */
    check &= S_IRWXG;
  } else { /* others */
    check &= S_IRWXO;
  }

  mode = FileNode_GetFlags(node);

  if ((mode & check) == check) {
    return 1; /* OK */
  }

  return 0; /* NG */
}

static int do_open(const char *path, FileDescriptor *fdescp,
		   int dir_access, int file_access)
{
  FileNode node, dir;
  FileDescriptor fdesc;

  node = find_node(path, NULL, &dir, NULL);

  if (!node)
    return -ENOENT;

  if (!check_permit(dir, get_user_id(), get_group_id(), dir_access))
    return -EPERM;

  if (!check_permit(node, get_user_id(), get_group_id(), file_access))
    return -EPERM;

  fdesc = FileDescriptor_Create(FILE_DESCRIPTOR_MODE_READONLY, NULL);

  if ((file_access & W_OK)) {
    FileDescriptor_SetMode(fdesc, FILE_DESCRIPTOR_MODE_READWRITE);
  }

  FileDescriptor_SetNode(fdesc, node);

  *fdescp = fdesc;

  return 0;
}

static int do_release(FileDescriptor fdesc)
{
  FileDescriptor_Destroy(fdesc);
  return 0;
}

int my_getattr(const char *path, struct stat *st)
{
  FileNode node;
  FileDescriptor fdesc;

  fprintf(stderr, "getattr: %s\n", path);

  if (do_open(path, &fdesc, R_OK, 0) < 0) {
    return -ENOENT;
  }

  node = FileDescriptor_GetNode(fdesc);
  get_stat(node, st);

  do_release(fdesc);

  return 0;
}

int my_readlink(const char *path, char *buf, size_t size)
{
  fprintf(stderr, "readlink: %s\n", path);
  return -ENOSYS;
}

int my_getdir(const char *path, fuse_dirh_t dirh, fuse_dirfil_t dirfil)
{
  fprintf(stderr, "getdir: %s\n", path);
  return -ENOSYS;
}

int my_mknod(const char *path, mode_t mode, dev_t dev)
{
  fprintf(stderr, "mknod: %s\n", path);
  return -ENOSYS;
}

int my_mkdir(const char *path, mode_t mode)
{
  FileNode node, dir;
  FileObject object;
  char *filename;

  fprintf(stderr, "mkdir: %s\n", path);

  node = find_node(path, NULL, &dir, &filename);

  if (node) {
    return -EEXIST;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  node = Dir_AddNode(dir, filename);
  FileNode_SetFlags(node, mode);
  FileNode_SetUserID(node, get_user_id());
  FileNode_SetGroupID(node, get_group_id());
  object = FileObject_Create(FileNode_GetPool(node), FILE_OBJECT_TYPE_DIR);
  FileNode_SetObject(node, object);

  return 0;
}

int my_unlink(const char *path)
{
  FileNode node, dir;

  fprintf(stderr, "unlink: %s\n", path);

  node = find_node(path, NULL, &dir, NULL);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  if (FileObject_GetType(FileNode_GetObject(node)) != FILE_OBJECT_TYPE_FILE)
    return -ENOENT;

  node = Dir_DelNode(dir, node);

  return 0;
}

int my_rmdir(const char *path)
{
  FileNode node, dir;

  fprintf(stderr, "rmdir: %s\n", path);

  node = find_node(path, NULL, &dir, NULL);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  if (FileObject_GetType(FileNode_GetObject(node)) != FILE_OBJECT_TYPE_DIR)
    return -ENOENT;

  if (!Dir_IsEmpty(node)) {
    return -ENOTEMPTY;
  }

  node = Dir_DelNode(dir, node);

  return 0;
}

int my_symlink(const char *path, const char *linkpath)
{
  fprintf(stderr, "symlink: %s\n", path);
  return my_link(path, linkpath);
}

int my_rename(const char *path, const char *newpath)
{
  FileNode node, dir;
  FileNode newnode, newdir;
  char *newname;

  fprintf(stderr, "rename: %s %s\n", path, newpath);

  node    = find_node(path,    NULL, &dir,    NULL);
  newnode = find_node(newpath, NULL, &newdir, &newname);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  node = Dir_ExtractNode(dir, node);
  if (!node) {
    return -ENOENT;
  }

  if (newnode) {
    switch (FileObject_GetType(FileNode_GetObject(newnode))) {
    case FILE_OBJECT_TYPE_DIR:
      /* newnode ǥ쥯ȥʤ顤newnodeФưư */
      /* Ǥ⤳̵ư */
      Dir_InsertNode(newnode, node);
      break;
    case FILE_OBJECT_TYPE_FILE:
      /* ե뤬Ǥˤʤ */
      node = Dir_DelNode(newdir, newnode);
      Dir_InsertNode(newdir, node);
      break;
    case FILE_OBJECT_TYPE_NONE:
    default:
      break;
    }
  } else {
    FileNode_NameWrite(node, newname);
    Dir_InsertNode(newdir, node);
  }

  return 0;
}

int my_link(const char *path, const char *linkpath)
{
  FileNode node, dir;
  FileNode newnode, newdir;
  FileObject object;
  char *newname;

  fprintf(stderr, "link: %s %s\n", path, linkpath);

  node    = find_node(path,     NULL, &dir,    NULL);
  newnode = find_node(linkpath, NULL, &newdir, &newname);

  if (!node) {
    return -ENOENT;
  }

  if (newnode) { /* ե뤬Ǥ¸ */
    return -EEXIST;
  }

  if (!check_permit(newdir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  newnode = Dir_AddNode(newdir, newname);
  FileNode_SetFlags(newnode, FileNode_GetFlags(node));
  FileNode_SetUserID(newnode, get_user_id());
  FileNode_SetGroupID(newnode, get_group_id());
  object = FileNode_GetObject(node);
  FileNode_SetObject(newnode, object);

  return 0;
}

int my_chmod(const char *path, mode_t mode)
{
  FileNode node, dir;

  fprintf(stderr, "chmod: %s\n", path);

  node = find_node(path, NULL, &dir, NULL);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  FileNode_SetFlags(node, mode);

  return 0;
}

int my_chown(const char *path, uid_t uid, gid_t gid)
{
  FileNode node, dir;

  fprintf(stderr, "chown: %s\n", path);

  node = find_node(path, NULL, &dir, NULL);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  if (uid != -1) FileNode_SetUserID(node, uid);
  if (gid != -1) FileNode_SetGroupID(node, gid);

  return 0;
}

int my_truncate(const char *path, off_t offset)
{
  FileNode node;
  FileDescriptor fdesc;

  fprintf(stderr, "truncate: %s\n", path);

  if (do_open(path, &fdesc, W_OK, W_OK)) {
    return -ENOENT;
  }

  node = FileDescriptor_GetNode(fdesc);
  FileObject_DataResize(FileNode_GetObject(node), offset);

  FileNode_SetModifiedTimeNow(node);

  do_release(fdesc);

  return 0;
}

int my_utime(const char *path, struct utimbuf *utbuf)
{
  fprintf(stderr, "utime: %s\n", path);
  return -ENOSYS;
}

int my_open(const char *path, struct fuse_file_info *info)
{
  FileDescriptor fdesc;
  int access, ret;

  fprintf(stderr, "open: %s %08x\n", path, info->flags);

  switch (info->flags & O_ACCMODE) {
  case O_RDONLY: access = R_OK; break;
  case O_WRONLY: access = W_OK; break;
  case O_RDWR:   access = R_OK | W_OK; break;
  default: return -EINVAL;
  }

  ret = do_open(path, &fdesc, access, access);
  if (ret < 0)
    return ret;

  info->fh = (int)fdesc;

  return 0;
}

int my_read(const char *path, char *buf, size_t size, off_t offset,
	    struct fuse_file_info *info)
{
  FileNode node;
  FileObject object;
  FileDescriptor fdesc;

  fprintf(stderr, "read: %s %d %d\n", path, size, (int)offset);

  fdesc = (FileDescriptor)(int)info->fh;

  node = FileDescriptor_GetNode(fdesc);
  object = FileNode_GetObject(node);

  FileNode_SetAccessTimeNow(node);

  return FileObject_DataRead(object, offset, size, buf);
}

int my_write(const char *path, const char *buf, size_t size, off_t offset,
	     struct fuse_file_info *info)
{
  FileNode node;
  FileObject object;
  FileDescriptor fdesc;

  fprintf(stderr, "write: %s %d %d\n", path, size, (int)offset);

  fdesc = (FileDescriptor)(int)info->fh;

  node = FileDescriptor_GetNode(fdesc);
  object = FileNode_GetObject(node);

  if (offset + size > FileObject_DataGetSize(object))
    FileObject_DataResize(object, offset + size);

  FileNode_SetModifiedTimeNow(node);

  return FileObject_DataWrite(object, offset, size, buf);
}

int my_statfs(const char *path, struct statvfs *stvfs)
{
  fprintf(stderr, "statfs: %s\n", path);

  memset(stvfs, 0, sizeof(*stvfs));
  stvfs->f_bsize = 0;
  stvfs->f_frsize = 0;
  stvfs->f_blocks = 0;
  stvfs->f_bfree = 0;
  stvfs->f_bavail = 0;
  stvfs->f_files = 0;
  stvfs->f_ffree = 0;
  stvfs->f_favail = 0;
  stvfs->f_namemax = 64;

  return 0;
}

int my_flush(const char *path, struct fuse_file_info *info)
{
  fprintf(stderr, "flush: %s\n", path);
  /* need */
  return 0;
}

int my_release(const char *path, struct fuse_file_info *info)
{
  FileDescriptor fdesc;
  fprintf(stderr, "release: %s\n", path);
  fdesc = (FileDescriptor)(int)info->fh;
  FileDescriptor_Destroy(fdesc);
  return 0;
}

int my_fsync(const char *path, int val, struct fuse_file_info *info)
{
  fprintf(stderr, "fsync: %s\n", path);
  /* need */
  return 0;
}

int my_setxattr(const char *path, const char *buf1, const char *buf2, size_t size, int val)
{
  return -ENOSYS;
}

int my_getxattr(const char *path, const char *buf1, char *buf2, size_t size)
{
  return -ENOSYS;
}

int my_listxattr(const char *path, char *buf, size_t size)
{
  return -ENOSYS;
}

int my_removexattr(const char *path, const char *buf)
{
  return -ENOSYS;
}

int my_opendir(const char *path, struct fuse_file_info *info)
{
  fprintf(stderr, "opendir: %s\n", path);
  return my_open(path, info);
}

int my_readdir(const char *path, void *buf, fuse_fill_dir_t filldir, off_t offset,
	       struct fuse_file_info *info)
{
  FileDescriptor fdesc;
  FileNode node, dir;
  struct stat stbuf;
  int i, number, ret, size;
  char *name;

  fprintf(stderr, "readdir: %s\n", path);

  fdesc = (FileDescriptor)(int)info->fh;

  dir = FileDescriptor_GetNode(fdesc);
  number = Dir_GetEntryNumber(dir);

  for (i = 0; i < number; i++) {
    node = Dir_GetNode(dir, i);
    if (!node)
      continue;

    get_stat(node, &stbuf);

    size = FileNode_NameGetSize(node);
    name = malloc(size);
    FileNode_NameRead(node, size, name);
    ret = filldir(buf, name, &stbuf, 0);
    free(name);
    if (ret == 1) {
      break;
    }
  }

  return 0;
}

int my_releasedir(const char *path, struct fuse_file_info *info)
{
  fprintf(stderr, "releasedir: %s\n", path);
  return my_release(path, info);
}

int my_fsyncdir(const char *path, int val, struct fuse_file_info *info)
{
  fprintf(stderr, "fsyncdir: %s\n", path);
  /* need */
  return 0;
}

void *my_init(struct fuse_conn_info *conn)
{
  struct fs_info *fs;
  FileObject object;
  mode_t mode;
  uid_t user_id;
  gid_t group_id;

  fprintf(stderr, "init:\n");

  fs = get_fs_info();

  if (fs->image_filename) {
    fprintf(stderr, "image file: %s\n", fs->image_filename);
  }

  mode     = 0777;
  user_id  = get_user_id();
  group_id = get_group_id();
  if (fs->mode) {
    mode     = fs->mode;
    user_id  = fs->user_id;
    group_id = fs->group_id;
  }

  /* ס */
#define BUFFER_SIZE (1024 * 1024 * 16)
  fs->pool = BlockPool_Create(malloc(BUFFER_SIZE), BUFFER_SIZE);

  /* ޥȥݥȤ */
  fs->mountpoint = FileNode_Create(fs->pool, "");
  FileNode_SetFlags(fs->mountpoint, 0755);
  FileNode_SetUserID(fs->mountpoint, 0);
  FileNode_SetGroupID(fs->mountpoint, 0);
  object = FileObject_Create(fs->pool, FILE_OBJECT_TYPE_DIR);
  FileNode_SetObject(fs->mountpoint, object);

  /* 롼ȥǥ쥯ȥ */
  fs->root = Dir_AddNode(fs->mountpoint, "");
  FileNode_SetFlags(fs->root, mode);
  FileNode_SetUserID(fs->root, user_id);
  FileNode_SetGroupID(fs->root, group_id);
  object = FileObject_Create(fs->pool, FILE_OBJECT_TYPE_DIR);
  FileNode_SetObject(fs->root, object);

  return fs;
}

void my_destroy(void *p)
{
  struct fs_info *fs;

  fprintf(stderr, "destroy:\n");

  fs = get_fs_info();

  BlockPool_Destroy(fs->pool);
  free(fs->pool);
  fs->pool = NULL;

  return;
}

int my_access(const char *path, int val)
{
  FileNode node, dir;
  unsigned int flags;

  fprintf(stderr, "access: %s %d\n", path, val);

  node = find_node(path, NULL, &dir, NULL);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), R_OK))
    return -EPERM;

  flags = FileNode_GetFlags(node);

  if ((val & R_OK) && !(flags & S_IRUSR)) return -1;
  if ((val & W_OK) && !(flags & S_IWUSR)) return -1;
  if ((val & X_OK) && !(flags & S_IXUSR)) return -1;

  return 0;
}

int my_create(const char *path, mode_t mode, struct fuse_file_info *info)
{
  FileNode node, dir;
  FileObject object;
  char *filename;

  fprintf(stderr, "create: %s %08x %08x\n", path, mode, info->flags);

  if (my_open(path, info) == 0) {
    return 0;
  }

  find_node(path, NULL, &dir, &filename);

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  node = Dir_AddNode(dir, filename);
  FileNode_SetFlags(node, mode);
  FileNode_SetUserID(node, get_user_id());
  FileNode_SetGroupID(node, get_group_id());
  object = FileObject_Create(FileNode_GetPool(node), FILE_OBJECT_TYPE_FILE);
  FileNode_SetObject(node, object);

  return my_open(path, info);
}

int my_ftruncate(const char *path, off_t offset, struct fuse_file_info *info)
{
  FileNode node;
  FileDescriptor fdesc;

  fprintf(stderr, "ftruncate: %s\n", path);

  fdesc = (FileDescriptor)(int)info->fh;
  node = FileDescriptor_GetNode(fdesc);
  FileObject_DataResize(FileNode_GetObject(node), offset);

  FileNode_SetModifiedTimeNow(node);

  return 0;
}

int my_fgetattr(const char *path, struct stat *st, struct fuse_file_info *info)
{
  FileNode node;
  FileDescriptor fdesc;

  fprintf(stderr, "fgetattr: %s\n", path);

  fdesc = (FileDescriptor)(int)info->fh;
  node = FileDescriptor_GetNode(fdesc);

  get_stat(node, st);

  return 0;
}

int my_lock(const char *path, struct fuse_file_info *info, int cmd,
	    struct flock *lock)
{
  fprintf(stderr, "lock: %s\n", path);
  /* need */
  return 0;
}

int my_utimens(const char *path, const struct timespec tv[2])
{
  FileNode node, dir;

  fprintf(stderr, "utimens: %s\n", path);

  node = find_node(path, NULL, &dir, NULL);

  if (!node) {
    return -ENOENT;
  }

  if (!check_permit(dir, get_user_id(), get_group_id(), W_OK))
    return -EPERM;

  FileNode_SetAccessTime(node, tv[0].tv_sec);
  FileNode_SetModifiedTime(node, tv[0].tv_sec);

  return 0;
}

int my_bmap(const char *path, size_t blocksize, uint64_t *idx)
{
  fprintf(stderr, "bmap: %s\n", path);
  return -ENOSYS;
}
